from redis_pal import RedisPal


def echo(arg):
    return arg


def setup():
    return "test", RedisPal()


def test_reprs():
    _, rp = setup()
    print(rp.__repr__())
    print(rp.__str__())
    assert rp.__str__() == rp.__repr__()


def test_int():
    key, rp = setup()
    # Integer
    inp = 1
    rp.set(key, inp)
    ans = rp.get(key)
    assert isinstance(ans, int)
    assert inp == ans


def test_float():
    key, rp = setup()
    # Float
    inp = 1.23
    rp.set(key, inp)
    ans = rp.get(key)
    assert isinstance(ans, float)
    assert inp == ans


def test_str():
    key, rp = setup()
    # String
    inp = "Test"
    rp.set(key, inp)
    ans = rp.get(key)
    assert isinstance(ans, str)
    assert str(inp) == ans


def test_bytes():
    key, rp = setup()
    # Bytes
    inp = b"Test"
    rp.set(key, inp)
    ans = rp.get(key)
    assert isinstance(ans, bytes)
    assert inp == ans


def test_func():
    key, rp = setup()
    # Function
    inp = echo
    rp.set(key, inp)
    ans = rp.get(key)
    assert ans(123) == 123


def test_numpy():
    import numpy as np
    key, rp = setup()
    # Numpy array
    inp = np.array([0, 1, 2, 3, 4])
    rp.set(key, inp)
    ans = rp.get(key)
    assert isinstance(ans, np.ndarray)
    assert inp[0] == ans[0]
    assert inp[1] == ans[1]
    assert inp[2] == ans[2]
    assert inp[3] == ans[3]
    assert inp[4] == ans[4]


def test_generator():
    key, rp = setup()
    inp = (num ** 2 for num in range(5))
    try:
        rp.set(key, inp)
        raise Exception("Could pickle a generator and shouldn't")
    except Exception:
        return


def test_unpickle_fail():
    _, rp = setup()
    try:
        rp._deserialize(b"\x80\x04}q\x00X\x01\x10\x00\x00aj\x01K{s.")
        raise Exception("Could unpickle corrupted data and shouldn't")
    except:
        return


def test_deserialize_bool():
    _, rp = setup()
    rp.set("test", True)
    assert rp.get("test")
    rp.set("test", False)
    assert not rp.get("test")


def test_force_dill():
    _, rp = setup()
    inp = compile("", "", "exec")
    enc = rp._serialize(inp)
    assert rp._deserialize(enc) == inp


def test_timestamp():
    from time import time
    start = time()
    key, rp = setup()
    # Integer
    inp = 1
    rp.set(key, inp)
    ans = rp.get_with_timestamp(key)
    assert isinstance(ans["value"], int)
    assert inp == ans["value"]
    assert (ans["last_modified"] - start) < 1


def test_zero_timestamp():
    _, rp = setup()
    ans = rp.get_with_timestamp("RANDOM_UNSET_KEY")
    assert ans["value"] is None
    assert ans["last_modified"] == 0


def test_None():
    _, rp = setup()
    inp = None
    ans = rp.get("RANDOM_UNSET_KEY")
    assert ans is None
    assert inp == ans



__all__ = ["SerializationError", "DeserializationError", "FatalError"]


class SerializationError(Exception):
    """
    Exception raised when
    serialization fails.
    """


class DeserializationError(Exception):
    """
    Exception raised when
    deserialization fails.
    """


class FatalError(Exception):
    """
    General purpose exception.
    """



__all__ = ["RedisPal"]

import pickle
from time import time
from typing import Union
import dill
import redis
from redis_pal.exceptions import SerializationError, DeserializationError


class RedisPal(redis.Redis):
    """
    RedisPal helps you store data on Redis
    without worrying about serialization
    stuff. Besides that, it also stores
    timestamps for last changes so you can
    check that whenever you want.

    You can use this as a key-value data
    structure using "get" and "set"
    methods.

    This class inherits from redis.Redis
    and the main difference is that it
    handles serialization and deserialization
    of objects so you can set anything to
    a key.
    """

    def __repr__(self) -> str:
        """
        Representation of this object
        """

        return "<RedisPal <pool={}>>".format(self.connection_pool)

    def __str__(self) -> str:
        """
        Representation of this object
        """

        return self.__repr__()

    @classmethod
    def _serialize(cls, o: object) -> bytes:
        try:
            return pickle.dumps(o)
        except Exception:
            try:
                return dill.dumps(o)
            except:
                raise SerializationError(
                    "Failed to serialize object {} of type {}".format(o, type(o)))

    @classmethod
    def _deserialize(cls, e: Union[str, int, float, bytes]) -> object:
        if e is None:
            return None
        try:
            return pickle.loads(e)
        except:
            try:
                return dill.loads(e)
            except:
                raise DeserializationError(
                    "Failed to deserialize {}".format(e))

    def set(self, key, value, *args, **kwargs) -> bool:
        _ser = self._serialize(value)
        _a = super(RedisPal, self).set(
            name="{}_timestamp".format(key), value=time(), *args, **kwargs)
        _b = super(RedisPal, self).set(
            name=key, value=_ser, *args, **kwargs
        )
        return all([_a, _b])

    def get(self, key, *args, **kwargs) -> object:
        return self._deserialize(super(RedisPal, self).get(name=key, *args, **kwargs))

    def get_with_timestamp(self, key, *args, **kwargs) -> dict:
        last_modified = super(RedisPal, self).get(
            name="{}_timestamp".format(key), *args, **kwargs)
        return {
            "value": self._deserialize(super(RedisPal, self).get(name=key, *args, **kwargs)),
            "last_modified": float(last_modified) if last_modified else 0
        }



from setuptools import setup, find_packages
from setuptools.extension import Extension
import subprocess
import sys


def install(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


try:
    from Cython.Build import cythonize
except:
    install("cython")
    from Cython.Build import cythonize

with open("README.md", "r") as fh:
    long_description = fh.read()

extensions = [
    Extension(
        "redis_pal.*",
        ["redis_pal/*.py"],
    ),
]

setup(
    name="redis_pal",
    version="0.1.2",
    license="GPL-3.0",
    description="Store things in Redis without worrying about types or anything, just do it!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    author="Gabriel Gazola Milan",
    author_email="gabriel.gazola@poli.ufrj.br",
    url="https://github.com/gabriel-milan/redis-pal",
    keywords=[
        "framework",
        "shared resources",
        "flexibility",
        "python",
        "online",
        "pipelines",
    ],
    install_requires=[
        "cython",
        "dill",
        "redis",
        "typing",
    ],
    extras_require={
        "test": [
            "numpy",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    ext_modules=cythonize(extensions, compiler_directives={
                          "language_level": "3"}),
)
